#FILE: random_sequence_reads.pl
#AUTH: Paul Stothard (paul.stothard@gmail.com)
#DATE: May 20, 2008
#VERS: 1.0

use warnings;
use strict;
use Data::Dumper;
use Getopt::Long;

my %options = (
    infile              => undef,
    outfile             => undef,
    n                   => undef,
    read_length         => 300,
    minimum_read_length => 300,
    discard             => undef
);

GetOptions(
    'i=s' => \$options{infile},
    'o=s' => \$options{outfile},
    'n=i' => \$options{n},
    'L=i' => \$options{read_length},
    'm=i' => \$options{minimum_read_length},
    'd'   => \$options{discard}
);

$options{n}                   = get_integer( $options{n} );
$options{read_length}         = get_integer( $options{read_length} );
$options{minimum_read_length} = get_integer( $options{minimum_read_length} );

if (   !( defined( $options{infile} ) )
    or !( defined( $options{outfile} ) )
    or !( defined( $options{n} ) )
    or !( defined( $options{read_length} ) )
    or !( defined( $options{minimum_read_length} ) ) )
{
    die( print_usage() . "\n" );
}

if ( ( $options{read_length} < 1 ) || ( $options{read_length} < 1 ) ) {
    die("-L and -m must be at least 1.");
}

open( my $INFILE, $options{infile} )
    or die("Cannot open file '$options{infile}': $!");

local $/ = ">";

#Count bases
my $bases                   = 0;
my $longest_sequence_length = 0;
while (<$INFILE>) {
    if ( $_ =~ m/[^\s>]/ ) {
        my @lines    = split( /\n/, $_ );
        my $title    = $lines[0];
        my $sequence = $_;
        $sequence =~ s/\Q$title\E//;
        $sequence =~ s/[^a-zA-Z]//g;
        $bases = $bases + length($sequence);
        if ( length($sequence) > $longest_sequence_length ) {
            $longest_sequence_length = length($sequence);
        }
    }
}
close($INFILE) or die("Cannot close file : $!");

if ( $options{minimum_read_length} > $longest_sequence_length ) {
    die("All sequences in $options{infile} are shorter than the minimum read length--unable to obtain reads."
    );
}

open( my $OUTFILE, '>', $options{outfile} )
    or die("Cannot open file '$options{outfile}': $!");

my $written_count = 0;
while ( $written_count < $options{n} ) {

    #An initial set of read positions is chosen, of size $options{n}.
    #Some of these will produce short reads that must be discarded.
    #Replacement read positions must be generated one at a time.
    my $positions;
    if ( $written_count == 0 ) {
        $positions = get_random_read_starts( $bases, $options{n} );
    }
    elsif ( $options{discard} ) {
        last;
    }
    else {
        $positions = get_random_read_starts( $bases, 1 );
    }

    #Write out sampled sequences
    my $current_end   = 0;
    my $current_start = 0;

    open( $INFILE, $options{infile} )
        or die("Cannot open file '$options{infile}': $!");
    while (<$INFILE>) {
        if ( $_ =~ m/[^\s>]/ ) {
            my @lines    = split( /\n/, $_ );
            my $title    = $lines[0];
            my $sequence = $_;
            $sequence =~ s/\Q$title\E//;
            $sequence =~ s/[^a-zA-Z]//g;
            $current_end = $current_end + length($sequence);
            for ( my $i = $current_start; $i < $current_end; $i++ ) {
                while ( $$positions[$i] > 0 ) {
                    my $read;
                    my $direct_strand = int( rand(2) );

                    if ($direct_strand) {
                        $read = substr( $sequence, $i - $current_start,
                            $options{read_length} );
                        if ( length($read) >= $options{minimum_read_length} )
                        {
                            print {$OUTFILE} ">$title"
                                . ";forward_strand;"
                                . ( $i - $current_start + 1 ) . "_to_"
                                . ( $i - $current_start + length($read) )
                                . "\n$read\n";
                            $written_count++;
                        }
                    }
                    else {
                        my $rev_comp = $sequence;
                        $rev_comp
                            =~ tr/gatcryswkmbdhvnGATCRYSWKMBDHVN/ctagyrswmkvhdbnCTAGYRSWMKVHDBN/;
                        $rev_comp = reverse($rev_comp);
                        $read
                            = substr( $rev_comp,
                            length($rev_comp) - 1 - ( $i - $current_start ),
                            $options{read_length} );
                        if ( length($read) >= $options{minimum_read_length} )
                        {
                            print {$OUTFILE} ">$title"
                                . ";reverse_strand;"
                                . ( $i - $current_start + 1 ) . "_to_"
                                . (
                                $i - $current_start + 1 - length($read) + 1 )
                                . "\n$read\n";
                            $written_count++;
                        }

                    }
                    $$positions[$i]--;
                }
            }
        }
        $current_start = $current_end;
    }
    close($INFILE) or die("Cannot close file : $!");
}

close($OUTFILE) or die("Cannot close file : $!");

print "Random sequences have been written to '$options{outfile}'\n";

sub get_random_read_starts {
    my $total_sequence_length = shift;
    my $n                     = shift;

    my $chosen_count = 0;
    my @chosen       = ();
    for ( my $i = 0; $i < $total_sequence_length; $i++ ) {
        $chosen[$i] = 0;
    }

    while ( $chosen_count < $n ) {
        my $random = int( rand($total_sequence_length) );
        $chosen[$random]++;
        $chosen_count++;
    }
    return \@chosen;
}

sub get_integer {
    my $value = shift;
    my $int   = undef;
    if ( ( defined($value) ) && ( $value =~ m/(\-*\d+)/ ) ) {
        $int = $1;
    }
    return $int;
}

sub print_usage {
    print <<BLOCK;
USAGE: perl random_sequence_reads.pl [-arguments]
 -i [FILE]     : file containing sequences in FASTA format (Required).
 -o [FILE]     : output file (Required).
 -n [INTEGER]  : number of reads to obtain (Required).
 -L [INTEGER]  : the desired read length (Optional. Default is 300).
 -m [INTEGER]  : the minimum read length (Optional. Default is 300).
 -d            : discard reads (Optional. Default is to resample short
                 reads).

perl random_sequence_reads.pl -i test/sample_in.fasta \
-o test/sample_out.fasta -n 10
BLOCK
}


